﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using FastTests;
using Raven.Client.Documents;
using Raven.Client.Documents.Indexes;
using Xunit;
using Xunit.Abstractions;

namespace SlowTests.Issues
{
    public class RavenDB_15608 : RavenTestBase
    {
        public RavenDB_15608(ITestOutputHelper output) : base(output)
        {
        }

        [Fact]
        public async Task ArtificalDocumentsMatchIndexEntries()
        {
            using (var store = GetDocumentStore())
            {
                await new SeasonHarvestEstimates_ByDeliverySite().ExecuteAsync(store);

                await PopulateSeasonVarietyGradeGroups(store);

                var year = "2021";
                var elementsCount = 15000;
                var siteCodes = Enumerable.Range(0, 150).Select(x => RandomString(5)).ToList();

                using (var session = store.OpenAsyncSession())
                {
                    var seasonVarietyGradeGroups = await session.Query<SeasonVarietyGradeGroup>().ToListAsync();

                    var seasonHarvestEstimate = new SeasonHarvestEstimate()
                    {
                        SeasonCode = year,
                        PropertyVarietyEstimates = new List<PropertyVarietyEstimate>()
                    };

                    var gradeGroups = seasonVarietyGradeGroups.Where(x => x.SeasonCode == year)
                        .Take(200).ToArray();

                    for (var i = 0; i < elementsCount; i++)
                    {
                        var gg = RandomElement(gradeGroups);

                        seasonHarvestEstimate.PropertyVarietyEstimates.Add(new PropertyVarietyEstimate
                        {
                            PropertyCode = i.ToString(),
                            NearestSiteCode = RandomElement(siteCodes),
                            DestinationSiteCode = RandomElement(siteCodes),
                            CommodityCode = gg.CommodityCode,
                            VarietyCode = gg.VarietyCode,
                            EstimatedHectares = _random.Next(20, 500)
                        });
                    }

                    await session.StoreAsync(seasonHarvestEstimate);

                    await session.SaveChangesAsync();
                }

                await AssertCounts(store);

                for (var i = 0; i < 30; i++)
                {
                    using (var session = store.OpenAsyncSession())
                    {
                        var randomToRemove = _random.Next(0, elementsCount);
                        session.Advanced.Patch<SeasonHarvestEstimate, PropertyVarietyEstimate>(SeasonHarvestEstimate.CalculateId(year), x => x.PropertyVarietyEstimates, x => x.RemoveAt(randomToRemove));
                        await session.SaveChangesAsync();
                        elementsCount--;
                    }

                    await AssertCounts(store);
                }
            }
        }

        private async Task AssertCounts(DocumentStore store)
        {
            WaitForIndexing(store);

            using (var session = store.OpenAsyncSession())
            {
                var indexCountLazy = session.Query<DeliverySiteEstimate, SeasonHarvestEstimates_ByDeliverySite>().CountLazilyAsync();
                var collectionCountLazy = session.Query<DeliverySiteEstimate>().CountLazilyAsync();

                var indexCount = await indexCountLazy.Value;
                var collectionCount = await collectionCountLazy.Value;

                Assert.Equal(indexCount, collectionCount);
            }
        }

        private static readonly Random _random = new Random();

        private static string RandomString(int length)
        {
            const string chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
            return new string(Enumerable.Repeat(chars, length)
                .Select(s => s[_random.Next(s.Length)]).ToArray());
        }

        private static T RandomElement<T>(ICollection<T> collection)
        {
            return collection.ElementAt(_random.Next(0, collection.Count));
        }

        private class SeasonVarietyGradeGroup
        {
            public static string CalculateId(string seasonCode, string commodityCode, string varietyCode) => $"seasonVarietyGradeGroup/{seasonCode}/{commodityCode}/{varietyCode}";
            public string Id => CalculateId(SeasonCode, CommodityCode, VarietyCode);

            public string SeasonCode { get; set; }
            public string CommodityCode { get; set; }
            public string VarietyCode { get; set; }
            public string GradeGroupCode { get; set; }
        }

        private class SeasonHarvestEstimate
        {
            public static string CalculateId(string seasonCode) => $"seasonHarvestEstimate/{seasonCode}";

            public string Id => CalculateId(SeasonCode);

            public string SeasonCode { get; set; }

            public List<PropertyVarietyEstimate> PropertyVarietyEstimates { get; set; }
        }

        private class PropertyVarietyEstimate
        {
            public string PropertyCode { get; set; }
            public string NearestSiteCode { get; set; }

            public string DestinationSiteCode { get; set; }

            public string CommodityCode { get; set; }
            public string VarietyCode { get; set; }
            public decimal EstimatedHectares { get; set; }
        }

        private class SeasonHarvestEstimates_ByDeliverySite : AbstractIndexCreationTask<SeasonHarvestEstimate, DeliverySiteEstimate>
        {
            public SeasonHarvestEstimates_ByDeliverySite()
            {
                Map = seasonHarvestEstimates => from seasonHarvestEstimate in seasonHarvestEstimates
                                                from propertyVarietyEstimate in seasonHarvestEstimate.PropertyVarietyEstimates
                                                select new DeliverySiteEstimate
                                                {
                                                    CommodityCode = propertyVarietyEstimate.CommodityCode,
                                                    DeliverySiteCode = propertyVarietyEstimate.DestinationSiteCode,
                                                    NearestSiteEstimates = new[]
                                                    {
                                                    new DeliverySiteEstimate.NearestSiteEstimate()
                                                    {
                                                        PropertyNearestSiteCode = propertyVarietyEstimate.NearestSiteCode,
                                                        Hectares = propertyVarietyEstimate.EstimatedHectares
                                                    }
                                                }
                                                };

                Reduce = results => from result in results
                                    group result by new { result.CommodityCode, result.DeliverySiteCode } into g
                                    select new DeliverySiteEstimate
                                    {
                                        CommodityCode = g.Key.CommodityCode,
                                        DeliverySiteCode = g.Key.DeliverySiteCode,
                                        NearestSiteEstimates = g.SelectMany(x => x.NearestSiteEstimates)
                                                                .GroupBy(x => x.PropertyNearestSiteCode)
                                                                .Select(gg => new DeliverySiteEstimate.NearestSiteEstimate()
                                                                {
                                                                    PropertyNearestSiteCode = gg.Key,
                                                                    Hectares = gg.Sum(x => x.Hectares)
                                                                }).ToArray()
                                    };

                OutputReduceToCollection = "DeliverySiteEstimates";
            }
        }

        private async Task PopulateSeasonVarietyGradeGroups(DocumentStore store)
        {
            using (var session = store.OpenAsyncSession())
            {
                #region StoreAsync
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "ALBUS", VarietyCode = "KIEVMUTANT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "BULOKE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "CHEBEC", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "CLIPPER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "COMMAN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "DASH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "DEVONGOLD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "DOOLUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FLAGSHIP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FORREST", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "GAIRD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "GALLEON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "GRANGE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "GROUT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "HANNAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "HARRINGTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "HENLEY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "HINDMA" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "KEEL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "KETCH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "MARITIME", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "MOLLOY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "MOONDYNE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "MORRELL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "NOYUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "SHANNON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "SKIPPER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "TALLON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "TANTANGAR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "TRIUMPH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "ULANDRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "UNICORN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "VLAMIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR155", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2030", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2034", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2080", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2104", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2109", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2110", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2147", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2312", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2315", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2321", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2452", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2473", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WABAR2482", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "WINDICH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "YAMBLA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "BARLEY", VarietyCode = "YARRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "43Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "44C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45C75", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y77", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "46C76", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "46Y20", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "46Y78", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "46Y81", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ARGYLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BAROSSA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "BUGLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "CHARLTON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "CHYB166", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "CLANCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "COUGAR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "DRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "DUNKELD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ECLIPSE", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "EMBLEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "EUREKA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "EYRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "FIGHTER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "FRONTIER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GRACE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GROUSE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "GT61", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HURRICANE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYDEN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYLITE200", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA42", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA474", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA50", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "HYOLA76", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "JARDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "JUNEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "KAROO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "LANTERN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "LIGHTNING", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "MALLEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "MALUKA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "MONTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "MUSTANG", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "MYSTIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "NARENDRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "OSCAR", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "OUTBACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PAC150", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PAC164", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PAC21063", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PC44C71", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PC44C72", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PC46C01", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PC46C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PC46C74", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "RAINBOW", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "RANGE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "RIVETTE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ROCKETCL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "ROTTNEST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SCOOP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SCORPION", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SENSATION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SHIRALEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SIGNAL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SIREN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SPECTRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "STORM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SUMMITT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP300TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP400", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP402CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP600", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP600TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURP603CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "SURPASS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TAIPAN", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TANAMI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TARCOOLA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TAWRIFFIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TI10", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TM8", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TRANBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TRIGOLD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "TRILOGY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "VIPER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "WARRIOR", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "WESBARKER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "WESBROOK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "X10N535R", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CANOLA", VarietyCode = "YICKADEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CERRYE", VarietyCode = "CEREALRYE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "AMETHYST", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "BARWON", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "CHICKPEAS", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "DESAVIC", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "DOOEN", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "GARNET", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "HEERA", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "HOWZAT", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "KANIVA", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "SEMSEN", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "SONA", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "CHIPEA", VarietyCode = "TYSON", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "ASCOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "BARCOOL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "CAIRO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "FIORD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FABEAN", VarietyCode = "ICARUS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "ALMA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "BUNDI", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "CELINE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "COLLEGIAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "COOGEE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "COOK", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "DERRIMUT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "DUNN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "DUNWA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "KING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "LAURA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "MAGNET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "MAITLAND", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "MORGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "PENNANT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "PRIMO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "SNOWPEAK", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "WHITE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "WHITEBRUNS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "WIRREGA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "FLDPEA", VarietyCode = "YARRUM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "CHITTICK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "MARRI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "MASON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "MOONAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "TALLERACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "UNICROP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "UNIWHITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "WANDOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "WODJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "AVON", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "BALLIDU", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "BANDICOOT", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "BETTONG", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "BRUSHER", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "BULBAN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "CASSIA", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "CONDAMINE", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "COOBA", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "DALE", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "DALYUP", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "DOLPHIN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "ECHIDNA", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "ESK", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "EURO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "FULMARK", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "GLIDER", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "HAKEA", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "HILL", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "IRWIN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "KALGAN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "KANGAROO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "KENT", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "LORT", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MARLOO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MASSIF", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MOORE", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MULGARA", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "MURRAY", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "NEEDILUP", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "NUMBAT", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "POSSUM", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "POTOROO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "QUOLL", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "STOUT", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "SWAN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "TOODYAY", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "TUNGOO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "VASSE", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WALLAROO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "WAND" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WEST", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WINJARDIE", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WINTAROO", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "WONGAN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "OATS", VarietyCode = "YILGARN", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "QUINOA", VarietyCode = "QUINOA", GradeGroupCode = "QUINOA" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "ABACUS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "COORONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "CURRENCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "EVEREST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "MUIR", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "OHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "TREAT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "TYALLA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "TRITIC", VarietyCode = "VENUS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "AMERY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ANGAS", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "AROONA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ARRIVATO", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "AXE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BABBLER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BANKS", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BARUNGA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BATAVIA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BENCUBBIN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BLANCO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BRAEWOOD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BRENNAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BTSCHOMBUR", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BUCKLEY", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BULLARING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "BUMPER", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CANNA", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CASTLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CATALINA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CHARA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "COCAMBA", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CONDOR", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CORRIGIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CUNDERDIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CUNNINGHAM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "CURRAWONG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "DARKAN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "DATATINE", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "DERRIMUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "DOLLARBIRD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "DRYSDALE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ELLISON", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ESTOC", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "EXCALIBUR", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "FALCON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "FANG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "FORD", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GABO", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GAMENYA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GILES", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GOLDMARK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GUARDIAN", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "GUTHA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "H45", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "H46", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "HARRISMITH", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "HERON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "JACUP", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "JANDAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "JANZCL", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "JITARNING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KALKA", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KAMILAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KARLGARIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KITE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KORD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KRICHAUF", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KUKRI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "LANCE", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "LANG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "LAWSON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "LINCOLN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "LORIKEET", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MACKELLAR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MADDEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MAROMBI", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MEERING", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MERCURYHY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MITRE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "MULGARA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "NAPAROO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "NYABING", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "PERENJORI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "PRESTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "PUGSLEY", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "QALBIS", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "RAVEN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "REES", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "REEVES", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ROSELLA", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "RUBY", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SCHOMBURGK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SCYTHE", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SHENTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SILVERSTAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "STYLET", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNBRI", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNECA", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNELG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNFIELD", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNLAND", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNLIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNMIST", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNSTAR", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNSTATE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUNVALE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TAILORBIRD", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TAKARI", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TAMAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TERN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TINCURRIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "TRIDENT", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "VULCAN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WAGIN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WARIGAL", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WEDIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WENTWORTH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WHISTLER", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WIALKI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WOLLAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WONGOONDY", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WORRAKATTA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "YALLAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "YARRALINKA", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "YOUNG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2015", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "ALBUS", VarietyCode = "KIEVMUTANT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "BULOKE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "CLIPPER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "COMMAN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "DASH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "DEVONGOLD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "DOOLUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FLAGSHIP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FLINDE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FORREST", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "GAIRD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "GRANGER", GradeGroupCode = "GRANGE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "GROUT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "HANNAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "HARRINGTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "HENLEY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "HINDMA" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "IGB1101", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "IGB1337", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "KEEL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "LATROB" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "MARITIME", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "MOLLOY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "NOYUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "SHANNON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "SKIPPER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "TALLON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "ULANDRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "UNICORN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "VLAMIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "WABAR2315", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "WABAR2537", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "WINDICH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "YAMBLA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "BARLEY", VarietyCode = "YARRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44Y26", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "44Y89", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45T01", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y77", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "46C76", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "46Y20", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "46Y78", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "46Y81", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ARGYLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BAROSSA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "BUGLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "CHARLTON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "CHYB166", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "CLANCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "COUGAR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "DG550", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "DRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "DUNKELD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ECLIPSE", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "FIGHTER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "FRONTIER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GRACE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GROUSE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GT42", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "GT61", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HURRICANE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA42", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA474", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA50", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA525", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA600", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA650", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA725", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "HYOLA971", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "IH51RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "IH52RR", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "JARDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "JUNEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "LIGHTNING", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "MALLEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "MUSTANG", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "OUTBACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "PC46C74", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "RIVETTE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ROCKETCL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "ROTTNEST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SCOOP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SCORPION", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SENSATION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SHIRALEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SIREN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SPECTRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "STORM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "SURP600TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TAIPAN", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TANAMI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TARCOOLA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TAWRIFFIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TI10", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TRANBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "VIPER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "WARRIOR", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "WESBARKER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "X10N535R", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CANOLA", VarietyCode = "YETNA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "AMETHYST", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "BARWON", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "DESAVIC", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "DOOEN", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "GARNET", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "HEERA", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "HOWZAT", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "KANIVA", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "SEMSEN", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "SONA", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "CHIPEA", VarietyCode = "TYSON", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "ASCOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "BARCOOL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "CAIRO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "FIORD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FABEAN", VarietyCode = "ICARUS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "COOGEE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "COOK", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "DUNN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "DUNWA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "LAURA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "MORGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "FLDPEA", VarietyCode = "YARRUM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "CHITTICK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "MARRI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "MASON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "TALLERACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "WANDOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "WODJIL", GradeGroupCode = "YELLOW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "WAND" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "QUINOA", VarietyCode = "QUINOA", GradeGroupCode = "QUINOA" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "CURRENCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "EVEREST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "FUSION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "AMERY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "AROONA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "AXE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BABBLER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BARUNGA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BATAVIA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BRAEWOOD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BREMER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BRENNAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BUCKLEY", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "BULLARING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CASTLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CATALINA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CHARA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "COCAMBA", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CUNNINGHAM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "CURRAWONG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "DARKAN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "DATATINE", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "DERRIMUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "DRYSDALE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ESTOC", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "EXCALIBUR", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "FALCON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "FANG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "GAMENYA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "GUTHA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "H45", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "H46", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "HARRISMITH", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "HYDRA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "IMPRESS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "JANZCL", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "JITARNING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KARLGARIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KITE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KORD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KUKRI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "LANG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "LINCOLN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "LORIKEET", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "MADDEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "MERCURYHY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "MITRE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "NAPAROO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "NYABING", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "PERENJORI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "PRESTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "REES", GradeGroupCode = "ASWT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "RUBY", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SCHOMBURGK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SCYTHE", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SILVERSTAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SPELT", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "STYLET", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SUNELG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SUNLIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SUNVALE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "TERN", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "TINCURRIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "TROJAN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WARIGAL", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WEDIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WENTWORTH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WIALKI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WORRAKATTA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "YOUNG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ZEN", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2016", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "CLIPPER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "COMMAN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "DASH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "DEVONGOLD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "DOOLUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FLAGSHIP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FLINDE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FORREST", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "GRANGER", GradeGroupCode = "GRANGE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "GROUT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "HANNAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "HARRINGTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "HENLEY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "HINDMA" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "IGB1337", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "KEEL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "LATROB" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "MARITIME", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "MOLLOY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "NOYUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "PLANET", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "ROSALIND", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SHANNON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SKIPPER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "SPARTACUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "TALLON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "ULANDRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "UNICORN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "WABAR2315", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "WABAR2537", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "WINDICH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "YAMBLA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "BARLEY", VarietyCode = "YARRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "3000TR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "43Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44Y26", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "44Y89", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45T01", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y77", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "46C76", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "46Y20", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "46Y78", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "46Y81", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ARGYLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BAROSSA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "BUGLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "CHARLTON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "CLANCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "COUGAR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "DG460", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "DG550", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "DRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ECLIPSE", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "FIGHTER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "FRONTIER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GRACE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GT42", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GT53", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "GT61", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HURRICANE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA474", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA50", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA525", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA600", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA650", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA725", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA970", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "HYOLA971", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "IH51RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "IH52RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "JARDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "LIGHTNING", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "MAKO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "MALLEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "MONOLA416", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "MUSTANG", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "RIVETTE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ROCKETCL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "ROTTNEST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SCOOP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SCORPION", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SIREN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SPECTRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "STORM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "SURP600TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TAIPAN", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TANAMI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TAWRIFFIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TI10", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TRANBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "TURBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "VIPER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "WARRIOR", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "X10N535R", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CANOLA", VarietyCode = "YETNA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CHIPEA", VarietyCode = "CHICKPEAS", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FABEAN", VarietyCode = "FIORD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FABEAN", VarietyCode = "SAMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "COOGEE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "DUNN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "KABULI", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "KABULI", VarietyCode = "GENESIS090", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "KABULI", VarietyCode = "KANIVA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "CHITTICK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "JURIEN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "MARRI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "MASON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "TALLERACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "UNIWHITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "WANDOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "WAND" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "CURRENCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "EVEREST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "FUSION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "BARLOO", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "BLANCHEFLE", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "CAPELLO", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "LANGUEDOC", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "MORAVA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "POPANY", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "RASINA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "VETCH", VarietyCode = "VOLGA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "AMERY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "AROONA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ARRIVATO", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "AXE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BABBLER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BRAEWOOD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BREMER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BRENNAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BUCKLEY", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "BULLARING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CASTLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CATALINA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CHARA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CUNNINGHAM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CURRAWONG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "CUTLASS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "DATATINE", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "DERRIMUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "DRYSDALE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ESTOC", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "FALCON", GradeGroupCode = "GENPUR" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "FANG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "GAMENYA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "GUTHA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "H45", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "H46", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "HARRISMITH", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "HYDRA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "IMPRESS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "JANDAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "JANZCL", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "JITARNING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KALKA", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KAMILAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KARLGARIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KITE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KORD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KUKRI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "LANG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "LINCOLN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "LORIKEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "MADDEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "MERCURYHY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "MITRE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "NAPAROO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "PERENJORI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "PRESTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "REES", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "RUBY", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SCEPTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SCHOMBURGK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SCYTHE", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SILVERSTAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "STYLET", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SUNELG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "TAMAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "TINCURRIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "TROJAN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WEDIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WOLLAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WORRAKATTA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "YALLAROI", GradeGroupCode = "DURUM" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "YOUNG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ZEN", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2017", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "ALBUS", VarietyCode = "KIEVMUTANT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "BULOKE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "COMMAN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "DASH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "DEVONGOLD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "FLAGSHIP", GradeGroupCode = "FLAGS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FLINDE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "GAIRD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "GRANGER", GradeGroupCode = "GRANGE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "HARRINGTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "HENLEY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "HINDMA" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "IGB1337", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "KEEL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "LATROB" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "MALTSTAR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "MOLLOY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "NOYUP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "PLANET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "ROSALIND", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "SHANNON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "SPARTACUS", GradeGroupCode = "SPART" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "BARLEY", VarietyCode = "YARRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "3000TR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44T02", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44Y26", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44Y89", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "44Y90", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45T01", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "45Y91", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "46Y20", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "46Y78", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ARGYLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "CHARLTON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "COUGAR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "DG460", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "DG550", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ECLIPSE", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "FIGHTER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "FRONTIER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GRACE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GT42", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GT53", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "GT61", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HURRICANE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA474", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA50", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA525", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA600", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA650", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA725", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "HYOLA970", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "IH51RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "IH52RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "JARDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "MAKO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "MONOLA416", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "MUSTANG", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "R3520", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "R5520P", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ROCKETCL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "ROTTNEST", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SCOOP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SCORPION", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SPECTRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "STORM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "T4510", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TANAMI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TAWRIFFIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TRANBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "TURBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "VIPER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CANOLA", VarietyCode = "YETNA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CERRYE", VarietyCode = "CEREALRYE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "CHICKPEAS", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "SLASHER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "CHIPEA", VarietyCode = "STRIKER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FABEAN", VarietyCode = "FIORD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FABEAN", VarietyCode = "SAMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "DUNN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "KABULI", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "KABULI", VarietyCode = "GENESIS090", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "KABULI", VarietyCode = "KANIVA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LENTIL", VarietyCode = "CASSAB", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LENTIL", VarietyCode = "DIGGER", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LENTIL", VarietyCode = "HURRICANE", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LENTIL", VarietyCode = "NORTHFIELD", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "CHITTICK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "JURIEN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "TALLERACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "WANDOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "DURACK", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "WAND" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "CURRENCY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "FUSION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "VETCH", VarietyCode = "LANGUEDOC", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "VETCH", VarietyCode = "MORAVA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "VETCH", VarietyCode = "RASINA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "VETCH", VarietyCode = "VOLGA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "AMERY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "AXE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BABBLER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BREMER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BUCKLEY", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "BULLARING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CASTLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CHIEF", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "CUTLASS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "DATATINE", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "DERRIMUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "DRYSDALE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ESTOC", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "FANG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "FORREST", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "GAMENYA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "GUTHA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "H45", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "H46", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "HARRISMITH", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "HAVOC", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "HYDRA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "IMPRESS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "JANZCL", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "JITARNING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KARLGARIN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KITE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KORD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KUKRI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "LANG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "LORIKEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "MADDEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "NINJA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "PASCAL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "PERENJORI", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "RUBY", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SCEPTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SUNELG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "TINCURRIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "TROJAN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "TUNGSTEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WEDIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WORRAKATTA", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "YOUNG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ZEN", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2018", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "BANKS", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "COMMAN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "DASH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "DEVONGOLD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FLINDE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "FORREST", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "GRANGER", GradeGroupCode = "GRANGE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "HENLEY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "IGB1337", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "LATROB" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "MALTSTAR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "MOLLOY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "PLANET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "ROSALIND", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "SHANNON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "SPARTACUS", GradeGroupCode = "SPART" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "BARLEY", VarietyCode = "YARRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "3000TR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "43Y92", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44T02", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y26", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y27", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y89", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "44Y90", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45T01", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "45Y91", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "46Y78", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "COUGAR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "CUTBURY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "DG408", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "DG460", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "DG550", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "DG670", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "ECLIPSE", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "FIGHTER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "FRONTIER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GRACE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GT42", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GT53", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "GT61", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA350", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA474", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA50", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA506", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA525", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA600", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA650", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA725", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA970", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "HYOLA971", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "IGNITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "IH51RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "IH52RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "JARDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "MAKO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "MONOLA416", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "R3520", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "R5520P", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "RIVETTE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "ROCKETCL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SCOOP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SCORPION", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SPECTRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "STORM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "T4510", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TANAMI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TRANBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TROPHY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "TURBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "VIPER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CANOLA", VarietyCode = "YETNA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CHIPEA", VarietyCode = "SLASHER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "CHIPEA", VarietyCode = "STRIKER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FABEAN", VarietyCode = "FIORD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FABEAN", VarietyCode = "SAMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "KABULI", VarietyCode = "GENESIS090", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LENTIL", VarietyCode = "BOLT", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LENTIL", VarietyCode = "CASSAB", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LENTIL", VarietyCode = "DIGGER", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LENTIL", VarietyCode = "HURRICANE", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LENTIL", VarietyCode = "NORTHFIELD", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "CHITTICK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "JURIEN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "TALLERACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "UNIWHITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "DURACK", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "ECHIDNA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "KOWARI", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "WAND" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "TRITIC", VarietyCode = "FUSION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "VETCH", VarietyCode = "CAPELLO", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "VETCH", VarietyCode = "LANGUEDOC", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "VETCH", VarietyCode = "RASINA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "VETCH", VarietyCode = "VOLGA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "AMERY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "AXE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BABBLER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BENNETT", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BREMER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "BULLARING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CASTLEROCK", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CHARA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CHIEF", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CURRAWONG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "CUTLASS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "DATATINE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "DERRIMUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ESTOC", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "FALCON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "FANG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "FORREST", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "GAMENYA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "GUTHA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "H45", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "H46", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "HARRISMITH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "HAVOC", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "HYDRA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "IMPRESS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "JITARNING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KARLGARIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KITE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KITTYHAWK", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KORD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KUKRI", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "LANG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "LINCOLN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "LONGSWORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "MADDEN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "MERCURYHY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "NINJA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "PASCAL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "PERENJORI", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "REES", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "RUBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SCEPTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SCYTHE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "STYLET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SUNELG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "TROJAN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "TUNGSTEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "WORRAKATTA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "YOUNG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ZEN", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2019", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BANKS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BOTTLER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BUFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "COMMAN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "DASH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "DEVONGOLD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FLINDE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "FORREST", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "GRANGER", GradeGroupCode = "GRANGE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "HENLEY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "IGB1337", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "LATROB" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "MALTSTAR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "MOLLOY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "PLANET", GradeGroupCode = "PLANET" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "ROSALIND", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SCOPIMI", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SHANNON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SPARIMI", GradeGroupCode = "SPART" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "SPARTACUS", GradeGroupCode = "SPART" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "BARLEY", VarietyCode = "YARRA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "3000TR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "43Y29", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "43Y92", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44C73", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44T02", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y26", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y27", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y89", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "44Y90", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45T01", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y91", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "45Y93", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "46Y78", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "530XT", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BANKER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "COUGAR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "CUTBURY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "DG408", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "DG460", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "DG550", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "DG670", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "ECLIPSE", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "FIGHTER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "FRONTIER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GRACE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GT42", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GT53", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "GT61", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA350", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA410XX", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA474", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA50", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA506", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA525", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA550", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA580", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA600", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA650", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA725", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA970", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "HYOLA971", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "IGNITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "IH51RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "IH52RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "JARDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "MAKO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "MONOLA416", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "R3520", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "R4020P", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "R5520P", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "RIVETTE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "ROCKETCL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SCOOP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SCORPION", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SPECTRUM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "STORM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "T4510", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TANAMI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TRANBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TROPHY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "TURBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "VIPER", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CANOLA", VarietyCode = "YETNA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "AMBAR", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "SLASHER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "CHIPEA", VarietyCode = "STRIKER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FABEAN", VarietyCode = "FIORD", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FABEAN", VarietyCode = "MARNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FABEAN", VarietyCode = "SAMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "BUTLER", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "KABULI", VarietyCode = "GENESIS090", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LENTIL", VarietyCode = "BOLT", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LENTIL", VarietyCode = "CASSAB", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LENTIL", VarietyCode = "DIGGER", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LENTIL", VarietyCode = "HALLMARK", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LENTIL", VarietyCode = "HURRICANE", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LENTIL", VarietyCode = "NORTHFIELD", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "CHITTICK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "JURIEN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "TALLERACK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "UNIWHITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "BILBY", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "DURACK", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "ECHIDNA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "KOWARI", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "MULGARA", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "WAND" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "WINJARDIE", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "TRITIC", VarietyCode = "FUSION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "VETCH", VarietyCode = "CAPELLO", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "VETCH", VarietyCode = "LANGUEDOC", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "VETCH", VarietyCode = "RASINA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "VETCH", VarietyCode = "VOLGA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "AMERY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "AXE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BABBLER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BENNETT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BREMER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "BULLARING", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CASTLEROCK", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CHARA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CHIEF", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CURRAWONG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "CUTLASS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "DATATINE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "DERRIMUT", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "DEVIL", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ESTOC", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "FALCON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "FANG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "FORREST", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "GAMENYA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "GUTHA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "H45", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "H46", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "HARRISMITH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "HAVOC", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "HYDRA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ILLABO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "IMPRESS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "JITARNING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KARLGARIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KINSEI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KITE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KITTYHAWK", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KORD", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KUKRI", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "LANG", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "LINCOLN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "LONGSWORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "MADDEN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "MERCURYHY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "NIGHTHAWK", GradeGroupCode = "" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "NINJA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "PASCAL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "PERENJORI", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "REES", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "RUBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SCEPTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SCYTHE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "STYLET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SUNELG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "TROJAN", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "TUNGSTEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "VIXEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "WORRAKATTA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "YOUNG", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ZEN", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2020", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "ALBUS", VarietyCode = "AMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "ALBUS", VarietyCode = "ANDROMEDA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "ALBUS", VarietyCode = "KIEVMUTANT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BANKS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BARQUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BASS", GradeGroupCode = "BASS" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BAUDIN", GradeGroupCode = "BAUDIN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BEECHER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BOTTLER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BUFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "BULOKE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "CAPSTAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "COMMANDER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "COMPASS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "DAMPIER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "FATHOM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "FITZGERALD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "FLEET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "FLINDERS", GradeGroupCode = "FLINDE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "FRANKLIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "GAIRDNER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "GRANGER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "HAMELIN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "HINDMARSH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "LATROBE", GradeGroupCode = "LATROB" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "LITMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "LOCKYER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "MALTSTAR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "MAXIMUS", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "MOBY", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "MUNDAH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "OCONNOR", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "ONSLOW", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "OXFORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "PLANET", GradeGroupCode = "PLANET" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "ROE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "ROSALIND", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "SCHOONER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "SCOPE", GradeGroupCode = "SCOPE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "SKIFF", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "SLOOP", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "SPARTACUS", GradeGroupCode = "SPART" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "STIRLING", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "URAMBIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "VLAMINGH", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "WESTMINSTE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "BARLEY", VarietyCode = "YAGAN", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "3000TR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "43C80", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "43Y23", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "43Y29", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "43Y85", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "43Y92", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44C79", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44T02", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y24", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y26", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y27", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y84CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y87", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y89", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "44Y90", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45T01", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45T03", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y21", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y22", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y25", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y28", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y82", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y86CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y88", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y91", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "45Y93", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "46Y83", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "530XT", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "ARCHER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "ATOMIC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "ATR409", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BANJO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BANKER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BARRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BEACON", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BONANZA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BONITO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BOOMER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "BRAVOTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "CARBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "COBBLER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "COBRA", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "CRUSHERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "CUTBURY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "DG408", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "DG460", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "DG550", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "DG670", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "FLINDERS", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "GEM", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "GT41", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "GT42", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "GT50", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "GT53", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HENTY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA350", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA404", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA410XX", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA450", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA502", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA505", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA506", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA525", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA540XC", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA550", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA555", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA559", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA571", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA575", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA577", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA580", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA600", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA601", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA650", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA656", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA725", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "HYOLA970", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "IGNITE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "IH30RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "IH50RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "IH51RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "IH52RR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "JACKPOT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "MAKO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "MARLIN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "PINNACLE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "R3520", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "R4022P", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "R5520P", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "RAPTOR", GradeGroupCode = "GMO" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "SAPPHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "SCADDAN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "SNAPPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "STINGRAY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "STUBBY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "STURT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "SURP404CL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "SURP501TT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "T3510", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "T4510", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TELFER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "THUMPER", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "THUNDERTT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TORNADO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TRIBUNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TRIDENT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TRIFECTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TROPHY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "TURBINE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "WAHOO", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CANOLA", VarietyCode = "YETNA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CERRYE", VarietyCode = "BEVY", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CERRYE", VarietyCode = "CEREALRYE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CERRYE", VarietyCode = "NIAB", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CERRYE", VarietyCode = "SLAV", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "AMBAR", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "CPPRIOR94", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "GENESIS510", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "GENESIS836", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "SLASHER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "STINGER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "CHIPEA", VarietyCode = "STRIKER", GradeGroupCode = "CHISTD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FABEAN", VarietyCode = "BENDOC", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FABEAN", VarietyCode = "FARAH", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FABEAN", VarietyCode = "FIESTA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FABEAN", VarietyCode = "MARNE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FABEAN", VarietyCode = "SAMIRA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "BUTLER", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "DUNDALE", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "GUNYAH", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "HELENA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "KASPA", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "PARAFIELD", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "PEARL", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "TWILIGHT", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "FLDPEA", VarietyCode = "WHARTON", GradeGroupCode = "MILLNG" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "KABULI", VarietyCode = "GARNET", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "KABULI", VarietyCode = "GENESIS090", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "KABULI", VarietyCode = "KANIVA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LENTIL", VarietyCode = "BOLT", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LENTIL", VarietyCode = "CASSAB", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LENTIL", VarietyCode = "DIGGER", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LENTIL", VarietyCode = "HALLMARK", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LENTIL", VarietyCode = "HURRICANE", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LENTIL", VarietyCode = "NORTHFIELD", GradeGroupCode = "LENTIL" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "BARLOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "BELARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "COROMUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "COYOTE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "DANJA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "GUNGURRU", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "GUNYIDI", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "ILLYARRIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "JENABILLUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "JURIEN", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "KALYA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "MANDELUP", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "MERRIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "MYALLIE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "QUILLINOCK", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "TANJIL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "WONGA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "YANDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "LUPINS", VarietyCode = "YORREL", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "BANNISTER", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "BILBY", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "CARROLUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "COOMALLO", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "DURACK", GradeGroupCode = "OAT2" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "HOTHAM", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "KOJONUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "KOWARI", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "MITIKA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "MORTLOCK", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "PALLINUP", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "WANDERING", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "WILLIAMS", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "OATS", VarietyCode = "YALLARA", GradeGroupCode = "OAT1" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "TRITIC", VarietyCode = "BERKSHIRE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "TRITIC", VarietyCode = "BOGONG", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "TRITIC", VarietyCode = "CREDIT", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "TRITIC", VarietyCode = "FUSION", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "TRITIC", VarietyCode = "SPEEDEE", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "TRITIC", VarietyCode = "TAHARA", GradeGroupCode = "STD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "VETCH", VarietyCode = "CAPELLO", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "VETCH", VarietyCode = "LANGUEDOC", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "VETCH", VarietyCode = "RASINA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "VETCH", VarietyCode = "RM4", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "VETCH", VarietyCode = "VOLGA", GradeGroupCode = "VETCH" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "2248", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ACCROC", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "AJANA", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ANNUELO", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ARRINO", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BAROOTAWON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BAXTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BENNETT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BINNU", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BLADE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BONNIEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BREMER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "BROOKTON", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CADOUX", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CALINGIRI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CAMM", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CARINYA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CARNAMAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CASCADES", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CATAPULT", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CHIEF", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CLEARSTL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "COBRA", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CORACK", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CORRELL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CURRAWONG", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "CUTLASS", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "DAGGER", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "DEVIL", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "EAGLEROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "EMUROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ENDURE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ENVOY", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ERADU", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ESPADA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "FORREST", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "FORTUNE", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "FRAME", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "GLADIUS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "GREGORY", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "GRENADE", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "HALBERD", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "HARPER", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "HAVOC", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "HYDRA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ILLABO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "IMPRESS", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "JANZ", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "JUSTICACL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "KALANNIE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "KATANA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "KINGROCK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "KINSEI", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "KUNJIN", GradeGroupCode = "SOFT" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "LEICHARDT", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "LONGSWORD", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "MACE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "MACHETE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "MAGENTA", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "NINJA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "PASCAL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "REVENUE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ROCKSTAR", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "SAPPHIRE", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "SCEPTER", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "SCOUT", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "SPEAR", GradeGroupCode = "ASW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "STILETTO", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "STYLET", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "SUNCO", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "SUPREME", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "TAMMARINRK", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "TROJAN", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "TUNGSTEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "VIXEN", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "WEDGETAIL", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "WESTONIA", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "WILGOYNE", GradeGroupCode = "FEED" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "WYALKATCHE", GradeGroupCode = "APWN" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "WYLAH", GradeGroupCode = "APW" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "YANDANOOKA", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "YITPI", GradeGroupCode = "HARD" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ZEN", GradeGroupCode = "NOODLE" });
                await session.StoreAsync(new SeasonVarietyGradeGroup { SeasonCode = "2021", CommodityCode = "WHEAT", VarietyCode = "ZIPPY", GradeGroupCode = "APW" });
                #endregion

                await session.SaveChangesAsync();
            }
        }

        private class DeliverySiteEstimate
        {
            //public string SeasonCode { get; set; }
            public string CommodityCode { get; set; }
            //public string GradeGroupCode { get; set; }
            public string DeliverySiteCode { get; set; }

            public NearestSiteEstimate[] NearestSiteEstimates { get; set; }

            public class NearestSiteEstimate
            {
                public string PropertyNearestSiteCode { get; set; }
                public decimal Hectares { get; set; }
            }

            public class ReduceOutputReferences
            {
                public string[] ReduceOutputs { get; set; }
            }
        }
    }
}
